@@ -41,7 +41,6 @@ import io.undertow.servlet.handlers.ServletRequestContext;
 import io.undertow.servlet.util.EmptyEnumeration;
 import io.undertow.servlet.util.IteratorEnumeration;
 import io.undertow.util.AttachmentKey;
-import io.undertow.util.CanonicalPathUtils;
 import io.undertow.util.DateUtils;
 import io.undertow.util.HeaderMap;
 import io.undertow.util.HeaderValues;
@@ -49,6 +48,24 @@ import io.undertow.util.Headers;
 import io.undertow.util.HttpString;
 import io.undertow.util.LocaleUtils;
 
+import javax.servlet.AsyncContext;
+import javax.servlet.DispatcherType;
+import javax.servlet.MultipartConfigElement;
+import javax.servlet.RequestDispatcher;
+import javax.servlet.ServletException;
+import javax.servlet.ServletInputStream;
+import javax.servlet.ServletRequest;
+import javax.servlet.ServletRequestWrapper;
+import javax.servlet.ServletResponse;
+import javax.servlet.ServletResponseWrapper;
+import javax.servlet.http.Cookie;
+import javax.servlet.http.HttpServletMapping;
+import javax.servlet.http.HttpServletRequest;
+import javax.servlet.http.HttpServletResponse;
+import javax.servlet.http.HttpSession;
+import javax.servlet.http.HttpUpgradeHandler;
+import javax.servlet.http.Part;
+import javax.servlet.http.PushBuilder;
 import java.io.BufferedReader;
 import java.io.IOException;
 import java.io.InputStreamReader;
@@ -73,29 +90,12 @@ import java.util.List;
 import java.util.Locale;
 import java.util.Map;
 import java.util.Set;
-import javax.servlet.AsyncContext;
-import javax.servlet.DispatcherType;
-import javax.servlet.MultipartConfigElement;
-import javax.servlet.RequestDispatcher;
-import javax.servlet.ServletException;
-import javax.servlet.ServletInputStream;
-import javax.servlet.ServletRequest;
-import javax.servlet.ServletRequestWrapper;
-import javax.servlet.ServletResponse;
-import javax.servlet.ServletResponseWrapper;
-import javax.servlet.http.Cookie;
-import javax.servlet.http.HttpServletMapping;
-import javax.servlet.http.HttpServletRequest;
-import javax.servlet.http.HttpServletResponse;
-import javax.servlet.http.HttpSession;
-import javax.servlet.http.HttpUpgradeHandler;
-import javax.servlet.http.Part;
-import javax.servlet.http.PushBuilder;
 
 /**
  * The http servlet request implementation. This class is not thread safe
  *
  * @author Stuart Douglas
+ * @author <a href="mailto:ropalka@redhat.com">Richard Opalka</a>
  */
 public final class HttpServletRequestImpl implements HttpServletRequest {
 
@@ -145,15 +145,17 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
     @Override
     public Cookie[] getCookies() {
         if (cookies == null) {
-            Map<String, io.undertow.server.handlers.Cookie> cookies = exchange.getRequestCookies();
-            if (cookies.isEmpty()) {
+            Iterable<io.undertow.server.handlers.Cookie> cookies = exchange.requestCookies();
+            int count = 0;
+            for (io.undertow.server.handlers.Cookie cookie : cookies) {
+                count++;
+            }
+            if (count == 0) {
                 return null;
             }
-            int count = cookies.size();
             Cookie[] value = new Cookie[count];
             int i = 0;
-            for (Map.Entry<String, io.undertow.server.handlers.Cookie> entry : cookies.entrySet()) {
-                io.undertow.server.handlers.Cookie cookie = entry.getValue();
+            for (io.undertow.server.handlers.Cookie cookie : cookies) {
                 try {
                     Cookie c = new Cookie(cookie.getName(), cookie.getValue());
                     if (cookie.getDomain() != null) {
@@ -230,38 +232,47 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
     public HttpServletMapping getHttpServletMapping() {
         ServletRequestContext src = exchange.getAttachment(ServletRequestContext.ATTACHMENT_KEY);
         ServletPathMatch match = src.getOriginalServletPathMatch();
-        if(getDispatcherType() == DispatcherType.FORWARD) {
+        final DispatcherType dispatcherType = getDispatcherType();
+        //UNDERTOW-1899 - ERROR is essentially forward operation
+        if(dispatcherType == DispatcherType.FORWARD || dispatcherType == DispatcherType.ERROR) {
             match = src.getServletPathMatch();
         }
         String matchValue;
-        switch (match.getMappingMatch()) {
-            case EXACT:
-                matchValue = match.getMatched();
-                if(matchValue.startsWith("/")) {
-                    matchValue = matchValue.substring(1);
-                }
-                break;
-            case DEFAULT:
-            case CONTEXT_ROOT:
-                matchValue = "";
-                break;
-            case PATH:
-                matchValue = match.getRemaining();
-                if(matchValue.startsWith("/")) {
-                    matchValue = matchValue.substring(1);
-                }
-                break;
-            case EXTENSION:
-                matchValue = match.getMatched().substring(0, match.getMatched().length() - match.getMatchString().length() + 1);
-                if(matchValue.startsWith("/")) {
-                    matchValue = matchValue.substring(1);
-                }
-                break;
-            default:
-                matchValue = match.getRemaining();
+        if (match != null) {
+            switch (match.getMappingMatch()) {
+                case EXACT:
+                    matchValue = match.getMatched();
+                    if (matchValue.startsWith("/")) {
+                        matchValue = matchValue.substring(1);
+                    }
+                    break;
+                case DEFAULT:
+                case CONTEXT_ROOT:
+                    matchValue = "";
+                    break;
+                case PATH:
+                    matchValue = match.getRemaining();
+                    if (matchValue == null) {
+                        matchValue = "";
+                    } else if (matchValue.startsWith("/")) {
+                        matchValue = matchValue.substring(1);
+                    }
+                    break;
+                case EXTENSION:
+                    String matched = match.getMatched();
+                    String matchString = match.getMatchString();
+                    int startIndex = matched.startsWith("/") ? 1 : 0;
+                    int endIndex = matched.length() - matchString.length() + 1;
+                    matchValue = matched.substring(startIndex, endIndex);
+                    break;
+                default:
+                    matchValue = match.getRemaining();
+            }
+            return new MappingImpl(matchValue, match.getMatchString(), match.getMappingMatch(), match.getServletChain().getManagedServlet().getServletInfo().getName());
+        }
+        else
+            return null;
         }
-        return new MappingImpl(matchValue, match.getMatchString(), match.getMappingMatch(), match.getServletChain().getManagedServlet().getServletInfo().getName());
-    }
 
     @Override
     public int getIntHeader(final String name) {
@@ -318,7 +329,7 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
             return false;
         }
         SecurityContext sc = exchange.getSecurityContext();
-        Account account = sc.getAuthenticatedAccount();
+        Account account = sc != null ? sc.getAuthenticatedAccount() : null;
         if (account == null) {
             return false;
         }
@@ -458,6 +469,10 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         }
 
         SecurityContext sc = exchange.getSecurityContext();
+        if (sc == null) {
+            throw UndertowServletMessages.MESSAGES.noSecurityContextAvailable();
+        }
+
         sc.setAuthenticationRequired();
         // TODO: this will set the status code and headers without going through any potential
         // wrappers, is this a problem?
@@ -482,7 +497,9 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
             throw UndertowServletMessages.MESSAGES.loginFailed();
         }
         SecurityContext sc = exchange.getSecurityContext();
-        if (sc.isAuthenticated()) {
+        if (sc == null) {
+            throw UndertowServletMessages.MESSAGES.noSecurityContextAvailable();
+        } else if (sc.isAuthenticated()) {
             throw UndertowServletMessages.MESSAGES.userAlreadyLoggedIn();
         }
         boolean login = false;
@@ -502,6 +519,9 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
     @Override
     public void logout() throws ServletException {
         SecurityContext sc = exchange.getSecurityContext();
+        if (sc == null) {
+            throw UndertowServletMessages.MESSAGES.noSecurityContextAvailable();
+        }
         sc.logout();
         if(servletContext.getDeployment().getDeploymentInfo().isInvalidateSessionOnLogout()) {
             HttpSession session = getSession(false);
@@ -682,10 +702,10 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
     }
 
     public void closeAndDrainRequest() throws IOException {
-        if(reader != null) {
+        if (reader != null) {
             reader.close();
         }
-        if(servletInputStream == null) {
+        if (servletInputStream == null) {
             servletInputStream = new ServletInputStreamImpl(this);
         }
         servletInputStream.close();
@@ -696,11 +716,15 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
      *
      */
     public void freeResources() throws IOException {
-        if(reader != null) {
-            reader.close();
-        }
-        if(servletInputStream != null) {
-            servletInputStream.close();
+        try {
+            if(reader != null) {
+                reader.close();
+            }
+            if(servletInputStream != null) {
+                servletInputStream.close();
+            }
+        } finally {
+            clearAttributes();
         }
     }
 
@@ -966,6 +990,9 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
 
     @Override
     public RequestDispatcher getRequestDispatcher(final String path) {
+        if (path == null) {
+            return null;
+        }
         String realPath;
         if (path.startsWith("/")) {
             realPath = path;
@@ -975,9 +1002,9 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
             if (lastSlash != -1) {
                 current = current.substring(0, lastSlash + 1);
             }
-            realPath = CanonicalPathUtils.canonicalize(current + path);
+            realPath = current + path;
         }
-        return new RequestDispatcherImpl(realPath, servletContext);
+        return servletContext.getRequestDispatcher(realPath);
     }
 
     @Override
@@ -1188,6 +1215,12 @@ public final class HttpServletRequestImpl implements HttpServletRequest {
         return "HttpServletRequestImpl [ " + getMethod() + ' ' + getRequestURI() + " ]";
     }
 
+    public void clearAttributes() {
+        if(attributes != null) {
+            this.attributes.clear();
+        }
+    }
+
     @Override
     public PushBuilder newPushBuilder() {
         if(exchange.getConnection().isPushSupported()) {
